{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Controlling Concurrency and Reproducibility with  the __order__ Clause"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The __order__ clause is used for controlling the parallel execution of  loop iterations for one or more loops that are associated with a directive.  It is specified with a clause argument and optional modifier.  The only supported argument, introduced in OpenMP 5.0, is the keyword  __concurrent__ which indicates that the loop iterations may execute  concurrently, including iterations in the same chunk per the loop schedule.  Because of the relaxed execution permitted with an __order(concurrent)__  clause, codes must not assume that any cross-iteration data dependences  would be preserved or that any two iterations may execute on the same thread."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The following example in this section demonstrates the use of  the __order(concurrent)__ clause, without any modifiers, for controlling  the parallel execution of loop iterations. The __order(concurrent)__ clause cannot be used for the second and third  __parallel__ __for__/__do__ constructs because of either having  data dependences or accessing threadprivate variables."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "//%compiler: clang\n",
    "//%cflags: -fopenmp\n",
    "\n",
    "/*\n",
    "* name: reproducible.1\n",
    "* type: C\n",
    "* version: omp_5.0\n",
    "*/\n",
    "#include <stdio.h>\n",
    "#include <omp.h>\n",
    "\n",
    "int main()\n",
    "{\n",
    "   const int n = 1000;\n",
    "   int v[n], u[n];\n",
    "   static int sum;\n",
    "   #pragma omp threadprivate(sum)\n",
    "\n",
    "   // no data dependences, so can execute concurrently\n",
    "   #pragma omp parallel for order(concurrent)\n",
    "   for (int i = 0; i < n; i++) {\n",
    "      u[i] = i;\n",
    "      v[i] = i;\n",
    "      v[i] += u[i] * u[i];\n",
    "   }\n",
    "\n",
    "   // with data dependences, so cannot execute iterations\n",
    "   // concurrently with the order(concurrent) clause\n",
    "   #pragma omp parallel for ordered\n",
    "   for (int i = 1; i < n; i++) {\n",
    "      v[i] += u[i] * u[i];\n",
    "      #pragma omp ordered\n",
    "   v[i] += v[i-1];\n",
    "   }\n",
    "\n",
    "   sum = 0;\n",
    "   // accessing a threadprivate variable, which would not be\n",
    "   // permitted if the order(concurrent) clause was present\n",
    "   #pragma omp parallel for copyin(sum)\n",
    "   for (int i = 0; i < n; i++) {\n",
    "      sum += v[i];\n",
    "   }\n",
    "\n",
    "   #pragma omp parallel\n",
    "   {\n",
    "      printf(\"sum = %d on thread %d\\n\", sum, omp_get_thread_num());\n",
    "   }\n",
    "\n",
    "   return 0;\n",
    "}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "!!%compiler: gfortran\n",
    "!!%cflags: -fopenmp\n",
    "\n",
    "! name: reproducible.1\n",
    "! type: F-free\n",
    "! version:    omp_5.0\n",
    "program main\n",
    "   use omp_lib\n",
    "   implicit none\n",
    "   integer, parameter :: n = 1000\n",
    "   integer :: v(n), u(n)\n",
    "   integer :: i\n",
    "   integer, save :: sum\n",
    "   !$omp threadprivate(sum)\n",
    "\n",
    "   !! no data dependences, so can execute concurrently\n",
    "   !$omp parallel do order(concurrent)\n",
    "   do i = 1, n\n",
    "      u(i) = i\n",
    "      v(i) = i\n",
    "      v(i) = v(i) + u(i) * u(i)\n",
    "   end do\n",
    "\n",
    "   !! with data dependences, so cannot execute iterations\n",
    "   !! concurrently with the order(concurrent) clause\n",
    "   !$omp parallel do ordered\n",
    "   do i = 2, n\n",
    "      v(i) = v(i) + u(i) * u(i)\n",
    "      !$omp ordered\n",
    "  v(i) = v(i) + v(i-1)\n",
    "      !$omp end ordered\n",
    "   end do\n",
    "\n",
    "   sum = 0\n",
    "   !! accessing a threadprivate variable, which would not be\n",
    "   !! permitted if the order(concurrent) clause was present\n",
    "   !$omp parallel do copyin(sum)\n",
    "   do i = 2, n\n",
    "      sum = sum + v(i)\n",
    "   end do\n",
    "\n",
    "   !$omp parallel\n",
    "      print *,\"sum = \",sum,\" on thread \", omp_get_thread_num()\n",
    "   !$omp end parallel\n",
    "\n",
    "end program"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Modifiers to the __order__ clause, introduced in OpenMP 5.1, may be  specified to control the reproducibility of the loop schedule for  the associated loop(s). A reproducible loop schedule will consistently  yield the same mapping of iterations to threads (or SIMD lanes) if the  directive name, loop schedule, iteration space, and binding region remain  the same. The __reproducible__ modifier indicates the loop schedule must  be reproducible, while the __unconstrained__ modifier indicates that  the loop schedule is not reproducible. If a modifier is not specified, then the __order__ clause does not affect  the reproducibility of the loop schedule."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The next example demonstrates the use of the __order(concurrent)__ clause  with modifiers for additionally controlling the reproducibility of a loop's  schedule. The two worksharing-loop constructs in the first __parallel__ construct specify that the loops have reproducible schedules, thus memory effects from iteration  _i_  from the first loop will be observable to iteration  _i_  in the second loop.  In the second __parallel__ construct, the __order__ clause does not  control reproducibility for the loop schedules. However, since both loops  specify the same static schedules, the schedules are reproducible and the  data dependences between the loops are preserved by the execution. In the third __parallel__ construct, the __order__ clause indicates  that the loops are not reproducible, overriding the default reproducibility prescribed by the specified static schedule. Consequentially,  the __nowait__ clause on the first worksharing-loop construct should not  be used to ensure that the data dependences are preserved by the execution."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "//%compiler: clang\n",
    "//%cflags: -fopenmp\n",
    "\n",
    "/*\n",
    "* name: reproducible.2\n",
    "* type: C\n",
    "* version: omp_5.1\n",
    "*/\n",
    "#include <stdio.h>\n",
    "\n",
    "int main()\n",
    "{\n",
    "   const int n = 1000;\n",
    "   int v[n], u[n];\n",
    "\n",
    "   #pragma omp parallel\n",
    "   {\n",
    "      // reproducible schedules are used for the following two constructs\n",
    "      #pragma omp for order(reproducible: concurrent) nowait\n",
    "      for (int i = 0; i < n; i++) {\n",
    "         u[i] = i;\n",
    "         v[i] = i;\n",
    "      }\n",
    "      #pragma omp for order(reproducible: concurrent)\n",
    "      for (int i = 0; i < n; i++) {\n",
    "         v[i]  += u[i] * u[i];\n",
    "      }\n",
    "   }\n",
    "\n",
    "   #pragma omp parallel\n",
    "   {\n",
    "      // static schedules preserve data dependences between the loops\n",
    "      #pragma omp for schedule(static) order(concurrent) nowait\n",
    "      for (int i = 0; i < n; i++) {\n",
    "         u[i] = i;\n",
    "         v[i] = i;\n",
    "      }\n",
    "      #pragma omp for schedule(static) order(concurrent)\n",
    "      for (int i = 0; i < n; i++) {\n",
    "         v[i]  += u[i] * u[i];\n",
    "      }\n",
    "   }\n",
    "\n",
    "   #pragma omp parallel\n",
    "   {\n",
    "      // the default reproducibility by the static schedule is not\n",
    "      // preserved due to the unconstrained order clause.\n",
    "      // use of nowait here could result in data race.\n",
    "      #pragma omp for schedule(static) order(unconstrained: concurrent)\n",
    "      for (int i = 0; i < n; i++) {\n",
    "         u[i] = i;\n",
    "         v[i] = i;\n",
    "      }\n",
    "      #pragma omp for schedule(static) order(unconstrained: concurrent)\n",
    "      for (int i = 0; i < n; i++) {\n",
    "         v[i]  += u[i] * u[i];\n",
    "      }\n",
    "   }\n",
    "\n",
    "   return 0;\n",
    "}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "!!%compiler: gfortran\n",
    "!!%cflags: -fopenmp\n",
    "\n",
    "! name: reproducible.2\n",
    "! type: F-free\n",
    "! version:    omp_5.1\n",
    "program main\n",
    "   implicit none\n",
    "   integer, parameter :: n = 1000\n",
    "   integer :: v(n), u(n)\n",
    "   integer :: i\n",
    "\n",
    "   !$omp parallel\n",
    "      !! reproducible schedules are used the following two constructs\n",
    "      !$omp do order(reproducible: concurrent) nowait\n",
    "      do i = 1, n\n",
    "         u(i) = i\n",
    "         v(i) = i\n",
    "      end do\n",
    "      !$omp do order(reproducible: concurrent)\n",
    "      do i = 1, n\n",
    "         v(i) = v(i) + u(i) * u(i)\n",
    "      end do\n",
    "   !$omp end parallel\n",
    "\n",
    "   !$omp parallel\n",
    "      !! static schedules preserve data dependences between the loops\n",
    "      !$omp do schedule(static) order(concurrent) nowait\n",
    "      do i = 1, n\n",
    "         u(i) = i\n",
    "         v(i) = i\n",
    "      end do\n",
    "      !$omp do schedule(static) order(concurrent)\n",
    "      do i = 1, n\n",
    "         v(i) = v(i) + u(i) * u(i)\n",
    "      end do\n",
    "   !$omp end parallel\n",
    "\n",
    "   !$omp parallel\n",
    "      !! the default reproducibility by the static schedule is not\n",
    "      !! preserved due to the unconstrained order clause.\n",
    "      !! use of nowait here could result in data race.\n",
    "      !$omp do schedule(static) order(unconstrained: concurrent)\n",
    "      do i = 1, n\n",
    "         u(i) = i\n",
    "         v(i) = i\n",
    "      end do\n",
    "      !$omp do schedule(static) order(unconstrained: concurrent)\n",
    "      do i = 1, n\n",
    "         v(i) = v(i) + u(i) * u(i)\n",
    "      end do\n",
    "   !$omp end parallel\n",
    "\n",
    "end program"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Native",
   "language": "native",
   "name": "native"
  },
  "language_info": {
   "file_extension": ".c",
   "mimetype": "text/plain",
   "name": "c"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}